@@ -304,6 +304,44 @@ class RangeLiteral(Expression):
         return [self.start, self.stop]
 
 
+class ArrayLiteral(Expression):
+    __slots__ = ("items",)
+
+    def __init__(self, token: TokenT, items: list[Expression]):
+        super().__init__(token)
+        self.items = items
+
+    def __eq__(self, other: object) -> bool:
+        return isinstance(other, ArrayLiteral) and self.items == other.items
+
+    def __str__(self) -> str:
+        return ", ".join(str(e) for e in self.items)
+
+    def __hash__(self) -> int:
+        return hash(tuple(self.items))
+
+    def __sizeof__(self) -> int:
+        return super().__sizeof__() + sys.getsizeof(self.items)
+
+    def evaluate(self, context: RenderContext) -> list[object]:
+        return [e.evaluate(context) for e in self.items]
+
+    async def evaluate_async(self, context: RenderContext) -> list[object]:
+        return [await e.evaluate_async(context) for e in self.items]
+
+    def children(self) -> list[Expression]:
+        return self.items
+
+    @staticmethod
+    def parse(stream: TokenStream, left: Expression) -> ArrayLiteral:
+        items: list[Expression] = [left]
+        while stream.current().type_ == TokenType.COMMA:
+            stream.next()  # ignore comma
+            items.append(parse_primitive(stream.current()))
+            stream.next()
+        return ArrayLiteral(left.token, items)
+
+
 class TemplateString(Expression):
     __slots__ = ("template",)
 
@@ -383,10 +421,6 @@ class Path(Expression):
             else:
                 self.path.append(segment)
 
-        if isinstance(self.path[0], Path):
-            # Flatten root segment
-            self.path = self.path[0].path + self.path[1:]
-
     def __str__(self) -> str:
         it = iter(self.path)
         buf = [str(next(it))]
@@ -495,8 +529,11 @@ class FilteredExpression(Expression):
 
     @staticmethod
     def parse(stream: TokenStream) -> FilteredExpression | TernaryFilteredExpression:
-        """Return a new FilteredExpression parsed from _tokens_."""
+        """Return a new FilteredExpression parsed from _stream_."""
         left = parse_primitive(stream.next())
+        if stream.current().type_ == TokenType.COMMA:
+            # Array literal syntax
+            left = ArrayLiteral.parse(stream, left)
         filters = Filter.parse(stream, delim=(TokenType.PIPE,))
 
         if is_token_type(stream.current(), TokenType.IF):
@@ -695,7 +732,7 @@ class Filter:
         try:
             return func(left, *positional_args, **keyword_args)
         except TypeError as err:
-            raise LiquidTypeError(f"{self.name}: {err}", token=self.token) from err
+            raise LiquidTypeError(str(err), token=self.token) from err
         except LiquidTypeError as err:
             err.token = self.token
             raise err
@@ -1442,33 +1479,14 @@ class LoopExpression(Expression):
             token=self.token,
         )
 
-    def _eval_int(self, expr: Expression | None, context: RenderContext) -> int | None:
-        if expr is None:
-            return None
-
-        val = expr.evaluate(context)
-        if not isinstance(val, int):
+    def _to_int(self, obj: object, *, token: TokenT) -> int:
+        try:
+            return to_int(obj)
+        except (ValueError, TypeError) as err:
             raise LiquidTypeError(
-                f"expected an integer, found {expr.__class__.__name__}",
-                token=expr.token,
-            )
-
-        return val
-
-    async def _eval_int_async(
-        self, expr: Expression | None, context: RenderContext
-    ) -> int | None:
-        if expr is None:
-            return None
-
-        val = await expr.evaluate_async(context)
-        if not isinstance(val, int):
-            raise LiquidTypeError(
-                f"expected an integer, found {expr.__class__.__name__}",
-                token=expr.token,
-            )
-
-        return val
+                f"expected an integer, found {obj.__class__.__name__}",
+                token=token,
+            ) from err
 
     def _slice(
         self,
@@ -1507,18 +1525,21 @@ class LoopExpression(Expression):
 
     def evaluate(self, context: RenderContext) -> tuple[Iterator[object], int]:
         it, length = self._to_iter(self.iterable.evaluate(context))
-        limit = self._eval_int(self.limit, context)
+        limit = (
+            self._to_int(self.limit.evaluate(context), token=self.limit.token)
+            if self.limit
+            else None
+        )
 
         match self.offset:
-            case StringLiteral(value=value):
+            case StringLiteral(value=value, token=token):
                 offset: str | int | None = value
                 if offset != "continue":
-                    raise LiquidSyntaxError(
-                        f"expected 'continue' or an integer, found '{offset}'",
-                        token=self.offset.token,
-                    )
+                    offset = self._to_int(offset, token=token)
+            case None:
+                offset = None
             case _offset:
-                offset = self._eval_int(_offset, context)
+                offset = self._to_int(_offset.evaluate(context), token=_offset.token)
 
         return self._slice(it, length, context, limit=limit, offset=offset)
 
@@ -1526,17 +1547,24 @@ class LoopExpression(Expression):
         self, context: RenderContext
     ) -> tuple[Iterator[object], int]:
         it, length = self._to_iter(await self.iterable.evaluate_async(context))
-        limit = await self._eval_int_async(self.limit, context)
+        limit = (
+            self._to_int(
+                await self.limit.evaluate_async(context), token=self.limit.token
+            )
+            if self.limit
+            else None
+        )
 
-        if isinstance(self.offset, StringLiteral):
-            offset: str | int | None = self.offset.evaluate(context)
+        if self.offset is None:
+            offset: str | int | None = None
+        elif isinstance(self.offset, StringLiteral):
+            offset = self.offset.evaluate(context)
             if offset != "continue":
-                raise LiquidSyntaxError(
-                    f"expected 'continue' or an integer, found '{offset}'",
-                    token=self.offset.token,
-                )
+                offset = self._to_int(offset, token=self.offset.token)
         else:
-            offset = await self._eval_int_async(self.offset, context)
+            offset = self._to_int(
+                await self.offset.evaluate_async(context), token=self.offset.token
+            )
 
         return self._slice(it, length, context, limit=limit, offset=offset)
 
@@ -1568,6 +1596,7 @@ class LoopExpression(Expression):
         reversed_ = False
         offset: Expression | None = None
         limit: Expression | None = None
+        cols: Expression | None = None
 
         while True:
             arg_token = stream.next()
@@ -1580,6 +1609,10 @@ class LoopExpression(Expression):
                         stream.expect_one_of(TokenType.COLON, TokenType.ASSIGN)
                         stream.next()
                         limit = parse_primitive(stream.next())
+                    case "cols":
+                        stream.expect_one_of(TokenType.COLON, TokenType.ASSIGN)
+                        stream.next()
+                        cols = parse_primitive(stream.next())
                     case "offset":
                         stream.expect_one_of(TokenType.COLON, TokenType.ASSIGN)
                         stream.next()
@@ -1615,7 +1648,7 @@ class LoopExpression(Expression):
             limit=limit,
             offset=offset,
             reversed_=reversed_,
-            cols=None,
+            cols=cols,
         )
 
 
@@ -1871,7 +1904,6 @@ def _contains(token: TokenT, left: object, right: object) -> bool:
     )
 
 
-# XXX: copied to avoid import issues
 def _to_liquid_string(val: Any, *, auto_escape: bool = False) -> str:
     """Stringify a Python object ready for output in a Liquid template."""
     if isinstance(val, str) or (auto_escape and hasattr(val, "__html__")):
